#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <pc.h>
#include <go32.h>
#include <dpmi.h>
#include <sys/farptr.h>
#include <grx20.h>
#include <keys.h>

#include "wave.h"

int ifd, ofd;
FILE *ifile;
int isize, istart, iblocks;
unsigned char *rbits;
unsigned char *lbits;
unsigned char *bits;
int end_offset;
char *fname;

int trim_start; /* ==0 means no trim - points to first kept*/
int trim_end; /* ==iblocks means no trim - points to first trimmed */

int bl, fg, bg, fgt, bgt, blt;

char *sbuf = 0;
int sbufcnt = 0;

int getbits(int b)
{
  if (!(bits[b/8] & (1 << (b%8))))
  {
    short buf[2352/2];
    int i, rmax=-65536, rmin=65536;
    int lmax=-65536, lmin=65536;
    fseek(ifile, b*2352+istart, 0);
    memset(buf, 0, 2352);
    fread(buf, 1, 2352, ifile);
    for (i=0; i<2352/2; i++)
    {
      if (lmin > buf[i])
	lmin = buf[i];
      if (lmax < buf[i])
	lmax = buf[i];
      i++;
      if (rmin > buf[i])
	rmin = buf[i];
      if (rmax < buf[i])
	rmax = buf[i];
    }
    i = rmax-rmin;
    rbits[b] = fls(i);
    i = lmax-lmin;
    lbits[b] = fls(i);
    bits[b/8] |= 1 << (b%8);
   }
  return b;
}

void plot(int y, int x, int b)
{
  int i, t=(b < trim_start || b >= trim_end);
  GrPlot(x, y, t ? blt : bl);
  for (i=0; i<17; i++)
  {
    if (t)
    {
      GrPlot(x, y-i-1, (i<lbits[b])?fgt:bgt);
      GrPlot(x, y+i+1, (i<rbits[b])?fgt:bgt);
    }
    else
    {
      GrPlot(x, y-i-1, (i<lbits[b])?fg:bg);
      GrPlot(x, y+i+1, (i<rbits[b])?fg:bg);
    }
  }
}

void box(int x, int y, int c)
{
  GrFilledBox(x*4+100, y*4+300, x*4+103, y*4+303, c);
}

void zoom(int bc, int x)
{
  int i, t, b;
  x *= GrMaxX()/4 - 80;
  for (b=bc-16; b<bc+16; b++)
  {
    if (b < 0 || b >= iblocks)
    {
      for (i=-17; i<17; i++)
	box(b-bc+x, i+1, 0);
    }
    else
    {
      t=(b < trim_start || b >= trim_end);
      box(b-bc+x, 0, t ? blt : bl);
      for (i=0; i<17; i++)
      {
	if (t)
	{
	  box(b-bc+x, -i-1, (i<lbits[b])?fgt:bgt);
	  box(b-bc+x, i+1, (i<rbits[b])?fgt:bgt);
	}
	else
	{
	  box(b-bc+x, -i-1, (i<lbits[b])?fg:bg);
	  box(b-bc+x, i+1, (i<rbits[b])?fg:bg);
	}
      }
    }
  }
}

void save(void)
{
  char buf[4];
  unsigned long l;
  unsigned short s;
  unsigned char *patch = (unsigned char *)malloc(2352 * trim_start + 8);
  int fd;

  memcpy(patch, "dj00", 4);
  *(unsigned long *)(patch+4) = trim_start * 2352 - 8;
  memcpy(patch+2352*trim_start, "data", 4);
  *(unsigned long *)(patch+2352*trim_start+4) = (trim_end - trim_start) * 2352;

  fd = open(fname, O_RDWR | O_BINARY);
  if (fd < 0)
    return;

  read(fd, buf, 4);
  if (memcmp(buf, "RIFF", 4))
    return;
  read(fd, &l, 4);
  read(fd, buf, 4);
  if (memcmp(buf, "WAVE", 0))
    return;
  while (1)
  {
    if (read(fd, buf, 4) < 4)
      return;
    else if (memcmp(buf, "data", 4) == 0)
    {
      lseek(fd, -4, SEEK_CUR);
      printf("Patching %d bytes at 0x%x\n", trim_start * 2352, tell(fd));
      if (trim_start)
	write(fd, patch, 2352 * trim_start + 8);
      if (trim_end < iblocks)
      {
	lseek(fd, (trim_end-trim_start)*2352, SEEK_CUR);
	*(unsigned long *)(patch+4) = (iblocks-trim_end) * 2352 - 8;
	write(fd, patch, 2352 * (iblocks-trim_end));
      }
      close(fd);
      return;
    }
    else
    {
      read(fd, &l, 4);
      lseek(fd, l, SEEK_CUR);
    }
  }
}

int
main(int argc, char **argv)
{
  __dpmi_regs r;
  int i, c, save_mode, ch;
  int done = 0;

  if (argc != 2)
  {
    printf("usage: trimwav infile outfile\n");
    exit(1);
  }

  ifd = open(argv[1], O_RDONLY | O_BINARY);
  ifile = fdopen(ifd, "rb");
  isize = setup_wave(ifd);
  istart = tell(ifd);
  iblocks = (isize+2351)/2352;

  fname = argv[1];

  rbits = (unsigned char *)malloc(iblocks);
  memset(rbits, 0, iblocks);
  lbits = (unsigned char *)malloc(iblocks);
  memset(lbits, 0, iblocks);
  bits = (unsigned char *)malloc((iblocks+7)/8);
  memset(bits, 0, (iblocks+7)/8);

  GrSetMode(GR_width_height_graphics, 800, 600);

  printf("%s -> %s", argv[1], argv[2]);
  printf("     a-s  k-l  esc to quit  Y to save");
  fflush(stdout);

  bg = 1;
  GrSetColor(bg, 0, 0, 128);
  bl = 2;
  GrSetColor(bl, 255, 255, 0);
  fg = 3;
  GrSetColor(fg, 0, 128, 255);

  bgt = 4;
  GrSetColor(bgt, 64, 0, 0);
  blt = 5;
  GrSetColor(blt, 128, 128, 0);
  fgt = 6;
  GrSetColor(fgt, 255, 0, 0);

  GrSetColor(7, 255, 255, 255);

  trim_start = 0;
  trim_end = iblocks;
  end_offset = iblocks - GrMaxX();

  getbits(0);
  for (i=0; i<GrMaxX(); i++)
  {
    getbits(i+1);
    if (!rbits[i] && !lbits[i] && (rbits[i+1] || lbits[i+1]))
    {
      trim_start = i+1;
      break;
    }
  }

  getbits(iblocks-1);
  for (i=iblocks-1; i>iblocks-GrMaxX(); i--)
  {
    getbits(i-1);
    if (!rbits[i] && !lbits[i] && (rbits[i-1] || lbits[i-1]))
    {
      trim_end = i;
      break;
    }
  }

  for (i=0; i<GrMaxX(); i++)
    plot(50, i, getbits(i));

  for (i=iblocks-GrMaxX(); i<iblocks; i++)
    plot(100, i-end_offset, getbits(i));

  zoom(trim_start,0);
  zoom(trim_end,1);

  while (!done) {
    int count=1;
    ch=getch();

    switch (ch) {
    case 27:
      done = 1;
      break;

    case 'y':
      save();
      done = 1;
      break;

    case 'A':
      count = 10;
    case '7':
    case 'a':
    case K_Home:
      while (count--) {
	if (trim_start) trim_start --;
	plot(50, trim_start, getbits(trim_start));
      }
      zoom(trim_start,0);
      break;

    case 'S':
      count = 10;
    case '8':
    case 's':
    case K_Up:
      while (count--) {
	trim_start ++;
	plot(50, trim_start-1, getbits(trim_start-1));
      }
      zoom(trim_start,0);
      break;

    case 'K':
      count = 10;
    case '2':
    case 'k':
    case K_Down:
      while (count--) {
	trim_end --;
	plot(100, trim_end-end_offset, getbits(trim_end));
      }
      zoom(trim_end,1);
      break;

    case 'L':
      count = 10;
    case '3':
    case 'l':
    case K_PageDown:
      while (count--) {
	if (trim_end < iblocks) trim_end ++;
	plot(100, trim_end-1-end_offset, getbits(trim_end-1));
      }
      zoom(trim_end,1);
      break;
    }
  }

  fclose(ifile);

  GrSetMode(GR_default_text);

  return 0;
}
