#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <pc.h>
#include <go32.h>
#include <dpmi.h>
#include <sys/farptr.h>
#include <grx20.h>
#include <keys.h>

#include "wave.h"

int ifd, ofd;
FILE *ifile;
int isize, istart, iblocks;
unsigned char *rbits;
unsigned char *lbits;
unsigned char *bits;
int end_offset;
char *wfname;

int trim_start; /* ==0 means no trim - points to first kept*/
int trim_end; /* ==iblocks means no trim - points to first trimmed */

int bl, fg, bg, fgt, bgt, blt;

char *sbuf = 0;
int sbufcnt = 0;

int getbits(int b)
{
  if (!(bits[b/8] & (1 << (b%8))))
  {
    short buf[2352/2];
    int i, rmax=-65536, rmin=65536;
    int lmax=-65536, lmin=65536;
    fseek(ifile, b*2352+istart, 0);
    memset(buf, 0, 2352);
    fread(buf, 1, 2352, ifile);
    for (i=0; i<2352/2; i++)
    {
      if (lmin > buf[i])
	lmin = buf[i];
      if (lmax < buf[i])
	lmax = buf[i];
      i++;
      if (rmin > buf[i])
	rmin = buf[i];
      if (rmax < buf[i])
	rmax = buf[i];
    }
    i = rmax-rmin;
    rbits[b] = fls(i);
    i = lmax-lmin;
    lbits[b] = fls(i);
    bits[b/8] |= 1 << (b%8);
   }
  return b;
}

void plot(int y, int x, int b)
{
  int i, t=(b < trim_start || b >= trim_end);
  GrPlot(x, y, t ? blt : bl);
  for (i=0; i<17; i++)
  {
    if (t)
    {
      GrPlot(x, y-i-1, (i<lbits[b])?fgt:bgt);
      GrPlot(x, y+i+1, (i<rbits[b])?fgt:bgt);
    }
    else
    {
      GrPlot(x, y-i-1, (i<lbits[b])?fg:bg);
      GrPlot(x, y+i+1, (i<rbits[b])?fg:bg);
    }
  }
}

void box(int x, int y, int c)
{
  GrFilledBox(x*4+100, y*4+300, x*4+103, y*4+303, c);
}

void zoom(int bc, int x)
{
  int i, t, b;
  x *= GrMaxX()/4 - 80;
  for (b=bc-16; b<bc+16; b++)
  {
    if (b < 0 || b >= iblocks)
    {
      for (i=-17; i<17; i++)
	box(b-bc+x, i+1, 0);
    }
    else
    {
      t=(b < trim_start || b >= trim_end);
      box(b-bc+x, 0, t ? blt : bl);
      for (i=0; i<17; i++)
      {
	if (t)
	{
	  box(b-bc+x, -i-1, (i<lbits[b])?fgt:bgt);
	  box(b-bc+x, i+1, (i<rbits[b])?fgt:bgt);
	}
	else
	{
	  box(b-bc+x, -i-1, (i<lbits[b])?fg:bg);
	  box(b-bc+x, i+1, (i<rbits[b])?fg:bg);
	}
      }
    }
  }
}

void save(void)
{
  int wfd;
  int bytes_left = (trim_end - trim_start) * 2352;
  int bytes_this;
  int wp=0, wpd=0, wb=0;
  int rp=0, rpd=0, rb=0;
  int bytes_per_pixel = bytes_left / (GrMaxX()-1);

  if (sbuf == 0)
  {
    sbufcnt = _go32_dpmi_remaining_physical_memory() / 2352;
    sbuf = (char *)sbrk(sbufcnt * 2352);
    if (sbuf == (char *)(-1))
    {
      GrSetMode(GR_default_text);
      printf("Cannot allocate %d bytes of memory\n", sbufcnt * 2352);
      exit(0);
    }
  }
  printf ("\n%d byte buffer\n", sbufcnt * 2352);

  GrFilledBox(0, 32, GrMaxX(), GrMaxY(), 0);

  GrFilledBox(0, 100, GrMaxX()-1, 140, blt);

  wfd = open(wfname, O_WRONLY | O_BINARY | O_CREAT | O_TRUNC, 0666);
  if (wfd < 0)
  {
      GrSetMode(GR_default_text);
      printf("Cannot open file %s\n", wfname);
      exit(0);
  }
  write_wave(wfd, bytes_left);

  lseek(ifd, istart + 2352 * trim_start, 0);

  while (bytes_left)
  {
    int i;
    char *cp;

    if (bytes_left < sbufcnt * 2352)
      bytes_this = bytes_left;
    else
      bytes_this = sbufcnt * 2352;

    i = bytes_this;
    cp = sbuf;
    while (i)
    {
      int r = i<16384 ? i : 16384;
      read(ifd, cp, r);
      i -= r;
      cp += r;
      rb += r;
      rpd = rb / bytes_per_pixel;
      while (rp < rpd)
      {
	GrLine(rp, 100, rp, 140, fgt);
	rp++;
      }
    }

    i = bytes_this;
    cp = sbuf;
    while (i)
    {
      int w = i<16384 ? i : 16384;
      write(wfd, cp, w);
      i -= w;
      cp += w;
      wb += w;
      wpd = wb / bytes_per_pixel;
      while (wp < wpd)
      {
	GrLine(wp, 100, wp, 140, fg);
	wp++;
      }
    }

    fsync(wfd);

    bytes_left -= bytes_this;
  }
  close(wfd);
  exit(0);
}

int
main(int argc, char **argv)
{
  __dpmi_regs r;
  int i, c, save_mode, ch;
  int done = 0;

  if (argc != 3)
  {
    printf("usage: trimwav infile outfile\n");
    exit(1);
  }

  ifd = open(argv[1], O_RDONLY | O_BINARY);
  ifile = fdopen(ifd, "rb");
  isize = setup_wave(ifd);
  istart = tell(ifd);
  iblocks = (isize+2351)/2352;

  wfname = argv[2];

  rbits = (unsigned char *)malloc(iblocks);
  memset(rbits, 0, iblocks);
  lbits = (unsigned char *)malloc(iblocks);
  memset(lbits, 0, iblocks);
  bits = (unsigned char *)malloc((iblocks+7)/8);
  memset(bits, 0, (iblocks+7)/8);

  GrSetMode(GR_width_height_graphics, 800, 600);

  printf("%s -> %s", argv[1], argv[2]);
  printf("     a-s  k-l  esc to quit  Y to save");
  fflush(stdout);

  bg = 1;
  GrSetColor(bg, 0, 0, 128);
  bl = 2;
  GrSetColor(bl, 255, 255, 0);
  fg = 3;
  GrSetColor(fg, 0, 128, 255);

  bgt = 4;
  GrSetColor(bgt, 64, 0, 0);
  blt = 5;
  GrSetColor(blt, 128, 128, 0);
  fgt = 6;
  GrSetColor(fgt, 255, 0, 0);

  GrSetColor(7, 255, 255, 255);

  trim_start = 0;
  trim_end = iblocks;
  end_offset = iblocks - GrMaxX();

  getbits(0);
  for (i=0; i<GrMaxX(); i++)
  {
    getbits(i+1);
    if (!rbits[i] && !lbits[i] && (rbits[i+1] || lbits[i+1]))
    {
      trim_start = i+1;
      break;
    }
  }

  getbits(iblocks-1);
  for (i=iblocks-1; i>iblocks-GrMaxX(); i--)
  {
    getbits(i-1);
    if (!rbits[i] && !lbits[i] && (rbits[i-1] || lbits[i-1]))
    {
      trim_end = i;
      break;
    }
  }

  for (i=0; i<GrMaxX(); i++)
    plot(50, i, getbits(i));

  for (i=iblocks-GrMaxX(); i<iblocks; i++)
    plot(100, i-end_offset, getbits(i));

  zoom(trim_start,0);
  zoom(trim_end,1);

  while (!done) {
    int count=1;
    ch=getch();

    switch (ch) {
    case 27:
      done = 1;
      break;

    case 'y':
      save();
      done = 1;
      break;

    case 'A':
      count = 10;
    case '7':
    case 'a':
    case K_Home:
      while (count--) {
	if (trim_start) trim_start --;
	plot(50, trim_start, getbits(trim_start));
      }
      zoom(trim_start,0);
      break;

    case 'S':
      count = 10;
    case '8':
    case 's':
    case K_Up:
      while (count--) {
	trim_start ++;
	plot(50, trim_start-1, getbits(trim_start-1));
      }
      zoom(trim_start,0);
      break;

    case 'K':
      count = 10;
    case '2':
    case 'k':
    case K_Down:
      while (count--) {
	trim_end --;
	plot(100, trim_end-end_offset, getbits(trim_end));
      }
      zoom(trim_end,1);
      break;

    case 'L':
      count = 10;
    case '3':
    case 'l':
    case K_PageDown:
      while (count--) {
	if (trim_end < iblocks) trim_end ++;
	plot(100, trim_end-1-end_offset, getbits(trim_end-1));
      }
      zoom(trim_end,1);
      break;
    }
  }

  fclose(ifile);

  GrSetMode(GR_default_text);

  return 0;
}
